/*	$Id: main.c,v 1.1.1.1 2007/04/06 22:48:25 tamentis Exp $ */

/*
 * Copyright (c) 2007 Bertrand Janin <tamentis@neopulsar.org>
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <arpa/inet.h>
#include <netinet/in.h>

#include <ctype.h>
#include <err.h>
#include <dirent.h>
#include <assert.h>
#include <errno.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <netdb.h>
#include <unistd.h>

#include "rethfu.h"

extern int h_errno;
extern char *__progname;
char *dbres;
char *tracks[MAX_TRACKS];
char *files[MAX_TRACKS];
unsigned int ntracks = 0;
unsigned int nfiles = 0;
char artist[MAX_ARTIST] = "";
char album[MAX_ALBUM] = "";
char genre[MAX_GENRE] = "";
char year[5];
char cat[32] = "";
char discid[9] = "";
unsigned int dbres_size;
unsigned int commit = 0;

size_t strlcpy(char *, const char *, size_t);

#define VERSION "$Id: main.c,v 1.1.1.1 2007/04/06 22:48:25 tamentis Exp $"

#define NROFREPS 28
char rep_from[NROFREPS] = "";
char rep_to[NROFREPS] =   "aaaaaaeeeeiiiioooooouuuuyyn";

/* convert_to_ascii - it takes utf strings and convert '' to 'o' i.e. */
char *
convert_to_ascii(char *org)
{
	char *dest;
	unsigned int i, j, len, offset = 0;

	len = strlen(org);

	dest = malloc(len + 1);

	for (i = 0; i < len; i++) {
		dest[i - offset] = org[i];
		for (j = 0; j < NROFREPS; j++) {
			if ( *(org + i + 0) == '' &&
			     *(org + i + 1) == rep_from[j]) {
				dest[i - offset] = rep_to[j];
				offset++;
				i++;
			}
		}
	}
	dest[i - offset] = '\0';

	return (dest);
}


#define GLIST_SIZE 148
static const char *glist[GLIST_SIZE] =
{
	"Blues",		/* 0 */
	"Classic Rock",
	"Country",
	"Dance",
	"Disco",
	"Funk",
	"Grunge",
	"Hip-Hop",
	"Jazz",
	"Metal",
	"New Age",		/* 10 */
	"Oldies",
	"Other",
	"Pop",
	"R&B",
	"Rap",
	"Reggae",
	"Rock",
	"Techno",
	"Industrial",
	"Alternative",		/* 20 */
	"Ska",
	"Death Metal",
	"Pranks",
	"Soundtrack",
	"Euro-Techno",
	"Ambient",
	"Trip-Hop",
	"Vocal",
	"Jazz+Funk",
	"Fusion",		/* 30 */
	"Trance",
	"Classical",
	"Instrumental",
	"Acid",
	"House",
	"Game",
	"Sound Clip",
	"Gospel",
	"Noise",
	"AlternRock",		/* 40 */
	"Bass",
	"Soul",
	"Punk",
	"Space",
	"Meditative",
	"Instrumental Pop",
	"Instrumental Rock",
	"Ethnic",
	"Gothic",
	"Darkwave",		/* 50 */
	"Techno-Industrial",
	"Electronic",
	"Pop-Folk",
	"Eurodance",
	"Dream",
	"Southern Rock",
	"Comedy",
	"Cult",
	"Gangsta",
	"Top 40",		/* 60 */
	"Christian Rap",
	"Pop/Funk",
	"Jungle",
	"Native American",
	"Cabaret",
	"New Wave",
	"Psychadelic",
	"Rave",
	"Showtunes",
	"Trailer",		/* 70 */
	"Lo-Fi",
	"Tribal",
	"Acid Punk",
	"Acid Jazz",
	"Polka",
	"Retro",
	"Musical",
	"Rock & Roll",
	"Hard Rock",
	/* winamp extentions */
	"Folk",			/* 80 */
	"Folk-Rock",
	"National Folk",
	"Swing",
	"Fast Fusion",
	"Bebob",
	"Latin",
	"Revival",
	"Celtic",
	"Bluegrass",
	"Avantgarde",		/* 90 */
	"Gothic Rock",
	"Progressive Rock",
	"Psychedelic Rock",
	"Symphonic Rock",
	"Slow Rock",
	"Big Band",
	"Chorus",
	"Easy Listening",
	"Acoustic",
	"Humour",		/* 100 */
	"Speech",
	"Chanson",
	"Opera",
	"Chamber Music",
	"Sonata",
	"Symphony",
	"Booty Bass",
	"Primus",
	"Porn Groove",
	"Satire",		/* 110 */
	"Slow Jam",
	"Club",
	"Tango",
	"Samba",
	"Folklore",
	"Ballad",
	"Power Ballad",
	"Rhythmic Soul",
	"Freestyle",
	"Duet",			/* 120 */
	"Punk Rock",
	"Drum Solo",
	"A capella",
	"Euro-House",
	"Dance Hall",
	"Goa",
	"Drum & Bass",
	"Club-House",
	"Hardcore",
	"Terror",		/* 130 */
	"Indie",
	"BritPop",
	"Negerpunk",
	"Polsk Punk",
	"Beat",
	"Christian Gangsta Rap",
	"Heavy Metal",
	"Black Metal",
	"Crossover",
	"Contemporary Christian",/* 140 */
	"Christian Rock ",
	"Merengue",
	"Salsa",
	"Trash Metal",
	"Anime",
	"JPop",
	"SynthPop"
};


/* get_gid - returns the genre short id from a char */
unsigned int
get_gid(char *gs)
{
	unsigned int i;

	for (i = 0; i < GLIST_SIZE; i++) {
		if (strcasecmp(gs, glist[i]) == 0) {
			return (i);
		}
	}

	return (200);
}


void
save_year(char *s)
{
	if ( *s != '\r') {
		strncpy(year, s, 4);
		year[4] = 0;
		//printf("Year: %s\n", year);
	}
}


void
save_genre(char *s)
{
	int i;

	for (i = 0; i < MAX_GENRE; i++) {
		if (s[i] == '\r') {
			genre[i] = 0;
			//printf("Genre: %s\n", genre);
			break;
		}
		genre[i] = s[i];
	}
}


void
save_track(char *s)
{
	char *endptr;
	char *buf;
	unsigned int tn, j;

	buf = malloc(TTRACK_MAX + 1);

	tn = (unsigned int)strtol(s, &endptr, 10) + 1;
	if (*endptr == '=') {
		endptr++;
		for (j = 0; j < TTRACK_MAX; j++) {
			if (*(endptr + j) == '\r') {
				*(buf + j) = 0;
				break;
			}
			*(buf + j) = *(endptr + j);
		}
		//printf("Track %02u: \"%s\"\n", tn, buf);
		tracks[tn - 1] = malloc(j + 1);
		strncpy(tracks[tn - 1], buf, j + 1);
		ntracks++;
	}
}


void
save_title(char *s)
{
	char *c;
	unsigned int i, j;
	
	c = artist;

	for (i = j = 0; i < TTITLE_MAX; i++, j++) {
		if (*(s + i) == '\r') {
			c[j] = 0;
			//printf("Artist: %s\n", artist);
			//printf("Album: %s\n", album);
			return;
		}
		if (strncmp(s + i, " / ", 3) == 0) {
			c[j] = 0;
			c = album;
			j = 0;
			i += 3;
		}
		c[j] = s[i];
	}


}

void
parse_response()
{
	unsigned int i;

	for (i = 0; i < dbres_size; i++) {
		if (strncmp(dbres + i, "\r\n\r\n", 4) == 0) {
			if (strncmp(dbres + i + 4, "501 ", 4) == 0)
				errx(1, CDDB_501);
			else if (strncmp(dbres + i + 4, "401 ", 4) == 0)
				errx(1, CDDB_401);
		} else if (*(dbres + i) == '\n') {
			if (strncmp(dbres + i + 1, "DTITLE=", 7) == 0) {
				save_title(dbres + i + 8);
			} else if (strncmp(dbres + i + 1, "DYEAR=", 6) == 0) {
				save_year(dbres + i + 7);
			} else if (strncmp(dbres + i + 1, "TTITLE", 6) == 0) {
				save_track(dbres + i + 7);
			} else if (strncmp(dbres + i + 1, "DGENRE=", 7) == 0) {
				save_genre(dbres + i + 8);
			}

		}
	}
}


ssize_t process_http(int sockfd, char *host, char *page, char *poststr)
{
	char sendline[MAXLINE + 1], recvline[MAXLINE + 1];
	char *buffer;
	unsigned int cur = 0;
	ssize_t n;

	snprintf(sendline, MAXSUB,
		 "POST %s HTTP/1.0\r\n"
		 "Host: %s\r\n"
		 "Content-type: application/x-www-form-urlencoded\r\n"
		 "Content-length: %d\r\n\r\n"
		 "%s", page, host, (int)strlen(poststr), poststr);

	buffer = malloc(BUFFER_SIZE);

	write(sockfd, sendline, strlen(sendline));
	while ((n = read(sockfd, recvline, MAXLINE)) > 0) {
		recvline[n] = '\0';
		strncpy(buffer + cur, recvline, n);
		cur += n;
	}

	dbres = malloc(cur);
	strncpy(dbres, buffer, cur);
	dbres_size = cur;
	free(buffer);

	//printf("%s", dbres);
	parse_response();

	return (n);

}

void
ask_freedb()
{
	int sockfd;
	char poststr[128];
	struct sockaddr_in servaddr;
	char str[50];
	struct hostent *hptr;
	char **pptr;
	char *hname = "freedb.freedb.org";
	char *page = "/~cddb/cddb.cgi";

	snprintf(poststr, 128, "cmd=cddb+read+%s+%s&hello=joe+my.host.com+rethfu+0.1&proto=6\r\n", cat, discid);

	if ((hptr = gethostbyname(hname)) == NULL) {
		errx(1, "gethostbyname error for host: %s: %s",
				hname, hstrerror(h_errno));
	}

	if (hptr->h_addrtype == AF_INET && (pptr = hptr->h_addr_list) != NULL) {
		inet_ntop(hptr->h_addrtype, *pptr, str, sizeof(str));
	} else {
		errx(1, "error calling inet_ntop");
	}

	sockfd = socket(AF_INET, SOCK_STREAM, 0);
	bzero(&servaddr, sizeof(servaddr));
	servaddr.sin_family = AF_INET;
	servaddr.sin_port = htons(80);
	inet_pton(AF_INET, str, &servaddr.sin_addr);

	connect(sockfd, (struct sockaddr *) & servaddr, sizeof(servaddr));
	process_http(sockfd, hname, page, poststr);
	close(sockfd);
}


void
find_files()
{
	DIR *dirp;
	struct dirent *dp;
	unsigned int i = 0;
	unsigned int namlen;

	dirp = opendir(".");
	if (dirp != NULL) {
		while ((dp = readdir(dirp)) != NULL) {
			/* booh to hidden */
			if (dp->d_name[0] == '.')
				continue;

			namlen = strlen(dp->d_name);
			/* booh to non-mp3 files */
			if (strncmp(dp->d_name + namlen - 4, ".mp3", 4) != 0)
				continue;

			files[i] = malloc(namlen + 1);
			strncpy(files[i], dp->d_name, namlen + 1);
			nfiles++;
			i++;
		}
		closedir(dirp);
	}

	/* reorder files in alphabetical order */
	char **ordered;
	unsigned int low = 0;
	char *s_low = NULL;
	unsigned int n;

	ordered = malloc(nfiles * sizeof(char*));
	s_low = files[0];

	for (n = 0; n < nfiles; n++) {
		for (i = 0; i < nfiles; i++) {
			if (files[i] == NULL)
				continue;
			if (strcmp(files[i], s_low) < 0) {
				s_low = files[i];
				low = i;
			}
		}
		ordered[n] = s_low;
		files[low] = NULL;
	
		/* find the next non-null entry */
		for (i = 0; i < nfiles; i++) {
			if (files[i] != NULL) {
				s_low = files[i];
				low = i;
				break;
			}
		}
	}

	/* put back pointers to **files */
	for (n = 0; n < nfiles; n++) {
		files[n] = ordered[n];
	}

	free(ordered);
}


int
findprog(char *prog, char *path, int allmatches)
{
	char *p, filename[MAXPATHLEN];
	int proglen, plen, rval = 0;
	struct stat sbuf;
	char *pathcpy;

	if ((path = strdup(path)) == NULL)
		errx(1, "Can't allocate memory.");
	pathcpy = path;

	proglen = strlen(prog);
	while ((p = strsep(&pathcpy, ":")) != NULL) {
		if (*p == '\0')
			p = ".";

		plen = strlen(p);
		while (p[plen-1] == '/')
			p[--plen] = '\0';	/* strip trailing '/' */

		if (plen + 1 + proglen >= sizeof(filename)) {
			warnx("%s/%s: %s", p, prog, strerror(ENAMETOOLONG));
			free(path);
			return (0);
		}

		snprintf(filename, sizeof(filename), "%s/%s", p, prog);
		if ((stat(filename, &sbuf) == 0) && S_ISREG(sbuf.st_mode) &&
		    access(filename, X_OK) == 0) {
			rval = 1;
			if (!allmatches) {
				free(path);
				return (rval);
			}
		}
	}
	(void)free(path);

	return (rval);
}


int
id3tag_available()
{
	char *path;
	int allmatches = 0, notfound = 0;

	if ((path = getenv("PATH")) == NULL)
		errx(1, "can't get $PATH from environment");

	/* To make access(2) do what we want */
	if (setgid(getegid()))
		err(1, "Can't set gid to %u", getegid());
	if (setuid(geteuid()))
		err(1, "Can't set uid to %u", geteuid());

	if (findprog("id3tag", path, allmatches) == 0)
		notfound++;

	return (!notfound);
}


void
usage()
{
	fprintf(stderr, "usage: %s [-drhDV] category/discid\n", __progname);
	exit(-1);
}


void
version()
{
	fprintf(stderr, "rethafu (rename those fuckers) - (c) 2007 Bertrand Janin\n");
	fprintf(stderr, "%s\n", VERSION);
	exit(-1);
}


int
main(int ac, char **av)
{
	int ch;
	int do_dump = 0;
	int do_fixid3 = 0;
	int do_rename = 0;
	int i;
	char cmdline[256];

	while ((ch = getopt(ac, av, "DdrhV")) != -1) {
		switch (ch) {
		case 'D':
			do_dump = 1;
			break;
		case 'd':
			do_fixid3 = 1;
			break;
		case 'r':
			do_rename = 1;
			break;
		case 'V':
			version();
		case 'h':
		default:
			usage();
		}
	}
	ac -= optind;
	av += optind;

	if (ac != 1 || strlen(av[0]) < 10)
		usage();

	/* find genre/discid */
	strlcpy(cat, av[0], 32);
	char *c = strchr(cat, '/');
	if (c == NULL)
		usage();
	*c = '\0';
	if (strlen(c + 1) != 8)
		usage();
	strncpy(discid, c + 1, 9);

	if (*cat == 0 || *discid == 0)
		usage();

	for (i = 0; i < strlen(cat); i++) {
		if (!isalpha(cat[i]))
			errx(1, ERROR_CAT);
	}

	for (i = 0; i < strlen(discid); i++) {
		if (!isxdigit(discid[i]))
			errx(1, ERROR_DID);
	}


	/* fetch freebd data from cat and discid */
	ask_freedb();

	if (do_dump) {
		printf("%s", dbres);
		exit(0);
	}

	if (*artist == 0)
		errx(1, "unable to get a correct artist name... sorry");
	if (*album == 0)
		errx(1, "unable to get this album name... sorry");
	if (*genre == 0)
		errx(1, "unable to get genre for this album... sorry");
	if (ntracks < 1)
		errx(1, "this album didn't have any track !");

	find_files();

	if (nfiles < 1)
		errx(1, "no mp3 file(s) in current directory.");
	if (ntracks != nfiles)
		errx(1, "not the same number of tracks (-f to force)");

	if (!do_fixid3 && !do_rename)
		errx(1, ERROR_NTD);

	if (do_fixid3) {
		if (!id3tag_available())
			errx(1, ERROR_ID3);
		for (i = 0; i < ntracks; i++) {
			unsigned int gid = get_gid(genre);

			printf("[fixid3] track %02u: %s\n", i + 1, tracks[i]);
			snprintf(cmdline, 256, ID3TAG_CMD, gid, i + 1, year, 
					artist, album, tracks[i], files[i]);
			system(cmdline);
		}
	}

	if (do_rename) {
		for (i = 0; i < ntracks; i++) {
			char nfn[256];
			char *tracks_a;
			int l, j;

			tracks_a = convert_to_ascii(tracks[i]);
			l = snprintf(nfn, 256, "%02u-%s.mp3", i + 1, tracks_a);
			free(tracks_a);
			for (j = 0; j < l; j++) {
				if (!isalnum(nfn[j]) && nfn[j] != '.' && nfn[j] != '-') 
					nfn[j] = '_';
				nfn[j] = tolower(nfn[j]);
			}
			printf("[rename] %s\n", nfn);
			if (rename(files[i], nfn) == -1)
				warn("rename issue");
		}
	}

	exit(0);
}

